'''
Created on Apr 8, 2014

@author: Jan Verhoeven

@copyright: MIT license, see http://opensource.org/licenses/MIT
'''
from ZmqReceiver import ZmqReceiver
from ZmqSender import ZmqSender
from threading import Thread
import logging

logger = logging.getLogger("zmqrpc")


# This class implements a simple message forwarding from a PUB/SUB connection to a
# REQ/REP connection.
class ZmqProxySub2Req(ZmqReceiver):
    # Note, at the moment username/password only protects the REQ-REP socket connection
    def __init__(self, zmq_sub_connect_addresses, zmq_req_connect_addresses, recreate_sockets_on_timeout_of_sec=600, username_sub=None, password_sub=None, username_req=None, password_req=None):
        ZmqReceiver.__init__(self, zmq_sub_connect_addresses=zmq_sub_connect_addresses, recreate_sockets_on_timeout_of_sec=recreate_sockets_on_timeout_of_sec, username=username_sub, password=password_sub)
        self.sender = ZmqSender(zmq_req_endpoints=zmq_req_connect_addresses, username=username_req, password=password_req)

    def handle_incoming_message(self, socket, message):
        # We don't care for the response, since we cannot pass it back via the pub socket or we got none from a pub socket
        try:
            self.sender.send(message, time_out_waiting_for_response_in_sec=60)
        except Exception as e:
            logger.error(e)
        return None


# This class implements a simple message forwarding from a PUB/SUB connection to another
# PUB/SUB connection. Could be used to aggregate messages into one end-point.
class ZmqProxySub2Pub(ZmqReceiver):
    # Note, at the moment username/password only protects the REQ-REP socket connection
    def __init__(self, zmq_sub_connect_addresses, zmq_pub_bind_address, recreate_sockets_on_timeout_of_sec=600, username_sub=None, password_sub=None, username_pub=None, password_pub=None):
        ZmqReceiver.__init__(self, zmq_sub_connect_addresses=zmq_sub_connect_addresses, recreate_sockets_on_timeout_of_sec=recreate_sockets_on_timeout_of_sec, username=username_sub, password=password_sub)
        self.sender = ZmqSender(zmq_pub_endpoint=zmq_pub_bind_address, username=username_pub, password=username_pub)

    def handle_incoming_message(self, socket, message):
        # We don't care for the response, since we cannot pass it back via the pub socket or we got none from a pub socket
        try:
            self.sender.send(message, time_out_waiting_for_response_in_sec=60)
        except Exception as e:
            logger.error(e)
        return None


# This class implements a simple message forwarding from a REQ/REP connection to a
# PUB/SUB connection.
class ZmqProxyRep2Pub(ZmqReceiver):
    # Note, at the moment username/password only protects the REQ-REP socket connection
    def __init__(self, zmq_rep_bind_address, zmq_pub_bind_address, recreate_sockets_on_timeout_of_sec=600, username_rep=None, password_rep=None, username_pub=None, password_pub=None):
        ZmqReceiver.__init__(self, zmq_rep_bind_address=zmq_rep_bind_address, recreate_sockets_on_timeout_of_sec=recreate_sockets_on_timeout_of_sec, username=username_rep, password=password_rep)
        self.sender = ZmqSender(zmq_req_endpoints=None, zmq_pub_endpoint=zmq_pub_bind_address, username=username_pub, password=password_pub)

    def handle_incoming_message(self, socket, message):
        try:
            self.sender.send(message, time_out_waiting_for_response_in_sec=60)
            # Pub socket does not provide response message, so return OK message
            return self.create_response_message(200, "OK", None)
        except Exception as e:
            return self.create_response_message(status_code=400, status_message="Error", response_message=e)


# This class implements a simple message forwarding from a REQ/REP connection to another
# REQ/REP connection.
class ZmqProxyRep2Req(ZmqReceiver):
    # Note, at the moment username/password only protects the REQ-REP socket connection
    def __init__(self, zmq_rep_bind_address, zmq_req_connect_addresses, recreate_sockets_on_timeout_of_sec=600, username_rep=None, password_rep=None, username_req=None, password_req=None):
        ZmqReceiver.__init__(self, zmq_rep_bind_address=zmq_rep_bind_address, recreate_sockets_on_timeout_of_sec=recreate_sockets_on_timeout_of_sec, username=username_rep, password=password_rep)
        self.sender = ZmqSender(zmq_req_endpoints=zmq_req_connect_addresses, username=username_req, password=password_req)

    def handle_incoming_message(self, socket, message):
        # Pass on the response from the forwarding socket.
        try:
            response_message = self.sender.send(message, time_out_waiting_for_response_in_sec=60)
            return self.create_response_message(status_code=200, status_message="OK", response_message=response_message)
        except Exception as e:
            return self.create_response_message(status_code=400, status_message="Error", response_message=e)


class ZmqProxyThread(Thread):
    def __init__(self):
        Thread.__init__(self)
        self.proxy = None

    def last_received_message(self):
        if self.proxy:
            return self.proxy.last_received_message
        return None

    def run(self):
        if self.proxy:
            self.proxy.run()

    def stop(self):
        if self.proxy:
            self.proxy.stop()


class ZmqProxySub2ReqThread(ZmqProxyThread):
    def __init__(self, zmq_sub_connect_addresses=None, zmq_req_connect_addresses=None, recreate_sockets_on_timeout_of_sec=600, username_sub=None, password_sub=None, username_req=None, password_req=None):
        ZmqProxyThread.__init__(self)
        self.proxy = ZmqProxySub2Req(zmq_sub_connect_addresses=zmq_sub_connect_addresses, zmq_req_connect_addresses=zmq_req_connect_addresses, recreate_sockets_on_timeout_of_sec=recreate_sockets_on_timeout_of_sec, username_sub=username_sub, password_sub=password_sub, username_req=username_req, password_req=password_req)


class ZmqProxySub2PubThread(ZmqProxyThread):
    def __init__(self, zmq_sub_connect_addresses=None, zmq_pub_bind_address=None, recreate_sockets_on_timeout_of_sec=600, username_sub=None, password_sub=None, username_pub=None, password_pub=None):
        ZmqProxyThread.__init__(self)
        self.proxy = ZmqProxySub2Pub(zmq_sub_connect_addresses=zmq_sub_connect_addresses, zmq_pub_bind_address=zmq_pub_bind_address, recreate_sockets_on_timeout_of_sec=recreate_sockets_on_timeout_of_sec, username_sub=username_sub, password_sub=password_sub, username_pub=username_pub, password_pub=password_pub)


class ZmqProxyRep2PubThread(ZmqProxyThread):
    def __init__(self, zmq_rep_bind_address=None, zmq_pub_bind_address=None, recreate_sockets_on_timeout_of_sec=600, username_rep=None, password_rep=None, username_pub=None, password_pub=None):
        ZmqProxyThread.__init__(self)
        self.proxy = ZmqProxyRep2Pub(zmq_rep_bind_address=zmq_rep_bind_address, zmq_pub_bind_address=zmq_pub_bind_address, recreate_sockets_on_timeout_of_sec=recreate_sockets_on_timeout_of_sec, username_rep=username_rep, password_rep=password_rep, username_pub=username_pub, password_pub=password_pub)


class ZmqProxyRep2ReqThread(ZmqProxyThread):
    def __init__(self, zmq_rep_bind_address=None, zmq_req_connect_addresses=None, recreate_sockets_on_timeout_of_sec=600, username_rep=None, password_rep=None, username_req=None, password_req=None):
        ZmqProxyThread.__init__(self)
        self.proxy = ZmqProxyRep2Req(zmq_rep_bind_address=zmq_rep_bind_address, zmq_req_connect_addresses=zmq_req_connect_addresses, recreate_sockets_on_timeout_of_sec=recreate_sockets_on_timeout_of_sec, username_rep=username_rep, password_rep=password_rep, username_req=username_req, password_req=password_req)


# This proxy class uses a 'hidden' pub/sub socket to buffer any messages from REP to REQ socket 
# in case the REQ socket is offline.
class ZmqBufferedProxyRep2ReqThread(ZmqProxyThread):
    def __init__(self, zmq_rep_bind_address=None, zmq_req_connect_addresses=None, buffered_pub_address="tcp://*:59878", buffered_sub_address="tcp://localhost:59878", recreate_sockets_on_timeout_of_sec=600, username_rep=None, password_rep=None, username_req=None, password_req=None):
        ZmqProxyThread.__init__(self)
        self.proxy1 = ZmqProxyRep2PubThread(zmq_rep_bind_address=zmq_rep_bind_address, zmq_pub_bind_address=buffered_pub_address, recreate_sockets_on_timeout_of_sec=100000, username_rep=username_rep, password_rep=password_rep)
        self.proxy2 = ZmqProxySub2ReqThread(zmq_sub_connect_addresses=[buffered_sub_address], zmq_req_connect_addresses=zmq_req_connect_addresses, recreate_sockets_on_timeout_of_sec=recreate_sockets_on_timeout_of_sec, username_req=username_req, password_req=password_req)

    def start(self):
        self.proxy1.start()
        self.proxy2.start()
        super(ZmqProxyThread, self).start()

    def stop(self):
        self.proxy1.stop()
        self.proxy2.stop()

    def join(self):
        self.proxy1.join()
        self.proxy2.join()
        super(ZmqProxyThread, self).join()



from setuptools import setup

setup(
    name='pyzmqrpc',
    packages=['zmqrpc'],
    version='1.0.1',
    include_package_data=True,
    description='A simple ZMQ RPC extension with JSON for message serialization',
    author='J Verhoeven',
    author_email='jan@visity.nl',
    url='https://github.com/jverhoeven/pyzmqrpc',
    download_url='https://github.com/jverhoeven/pyzmqrpc/tarball/1.0.0',
    keywords=['zeromq', 'rpc', 'pyzmq'],
    install_requires=["pyzmq>=14.1.0"],
    classifiers=[]
)



'''
Created on Apr 8, 2014

@author: Jan Verhoeven

@copyright: MIT license, see http://opensource.org/licenses/MIT
'''
import zmq
import zmq.auth
from zmq.auth.thread import ThreadAuthenticator
import json
import logging
import time
from threading import Thread

logger = logging.getLogger("zmqrpc")


# A ZmqReceiver class will listen on a REP or SUB socket for messages and will invoke a 'HandleIncomingMessage'
# method to process it. Subclasses should override that. A response must be implemented for REP sockets, but
# is useless for SUB sockets.
class ZmqReceiver():
    def __init__(self, zmq_rep_bind_address=None, zmq_sub_connect_addresses=None, recreate_sockets_on_timeout_of_sec=600, username=None, password=None):
        self.context = zmq.Context()
        self.auth = None
        self.last_received_message = None
        self.is_running = False
        self.thread = None
        self.zmq_rep_bind_address = zmq_rep_bind_address
        self.zmq_sub_connect_addresses = zmq_sub_connect_addresses
        self.recreate_sockets_on_timeout_of_sec = recreate_sockets_on_timeout_of_sec
        if username is not None and password is not None:
            # Start an authenticator for this context.
            # Does not work on PUB/SUB as far as I (probably because the more secure solutions
            # require two way communication as well)
            self.auth = ThreadAuthenticator(self.context)
            self.auth.start()
            # Instruct authenticator to handle PLAIN requests
            self.auth.configure_plain(domain='*', passwords={username: password})

    # May take up to 60 seconds to actually stop since poller has timeout of 60 seconds
    def stop(self):
        self.is_running = False
        logger.info("Closing pub and sub sockets...")
        if self.auth is not None:
            self.auth.stop()

    def run(self):
        self.is_running = True
        rep_socket = None
        sub_socket = None
        last_received_sample_timestamp = time.time()

        poller = zmq.Poller()
        create_new_socket_and_poller = True
        while self.is_running:
            # In case of any socket and poller errors: recreate them. Seems to be the only right way in ZeroMQ.
            if create_new_socket_and_poller:
                logger.debug("Creating new socket and poller")
                if rep_socket is not None:
                    poller.unregister(rep_socket)
                    rep_socket.setsockopt(zmq.LINGER, 0)
                    rep_socket.unbind(self.zmq_rep_bind_address)
                    rep_socket.close()
                    rep_socket = None
                    # For some reason, the req_socket binding lasts a bit longer than a couple of milliseconds, so give it a second or so to disconnect
                    time.sleep(1)
                if sub_socket is not None:
                    poller.unregister(sub_socket)
                    sub_socket.setsockopt(zmq.LINGER, 0)
                    for address in self.zmq_sub_connect_addresses:
                        sub_socket.disconnect(address)
                    sub_socket.close()
                    while not sub_socket.closed:
                        time.sleep(1)
                    sub_socket = None
                # Don't try to catch the exception in the following block: If it fails, it should be noted to the calling function via standard exceptions:
                if self.zmq_rep_bind_address is not None:
                    rep_socket = self.context.socket(zmq.REP)
                    if self.auth:
                        rep_socket.plain_server = True
                    rep_socket.bind(self.zmq_rep_bind_address)
                    poller.register(rep_socket, zmq.POLLIN)
                if self.zmq_sub_connect_addresses is not None:
                    sub_socket = self.context.socket(zmq.SUB)
                    sub_socket.setsockopt(zmq.SUBSCRIBE, '')
                    for address in self.zmq_sub_connect_addresses:
                        sub_socket.connect(address)
                    poller.register(sub_socket, zmq.POLLIN)
                create_new_socket_and_poller = False
                last_received_sample_timestamp = time.time()

            socks = dict(poller.poll(1000))
            incoming_message = None
            socket = None
            if socks.get(rep_socket) == zmq.POLLIN:
                last_received_sample_timestamp = time.time()
                socket = rep_socket
            elif socks.get(sub_socket) == zmq.POLLIN:
                last_received_sample_timestamp = time.time()
                socket = sub_socket
            else:
                # This code exists due to a bug in the library for ARM devices where one subscription can block another
                if time.time() > last_received_sample_timestamp + self.recreate_sockets_on_timeout_of_sec:
                    logger.debug("No new messages received since {0} seconds. Re-establishing subcriptions".format(self.recreate_sockets_on_timeout_of_sec))
                    create_new_socket_and_poller = True

            if socket is not None:
                try:
                    incoming_message = socket.recv()
                    self.last_received_message = incoming_message
                except Exception as e:
                    create_new_socket_and_poller = True
                    logger.error(e)
                else:
                    response_message = self.handle_incoming_message(socket, incoming_message)

            if rep_socket and socket == rep_socket:
                # Only send answers for reply sockets
                try:
                    if response_message is not None:
                        rep_socket.send(response_message)
                except Exception as e:
                    logger.error("Could not send message over socket. Recreating all sockets {0}".format(e))
                    create_new_socket_and_poller = True

        if rep_socket:
            rep_socket.close()
        if sub_socket:
            sub_socket.close()

    def create_response_message(self, status_code, status_message, response_message):
        if response_message is not None:
            return json.dumps({"status_code": status_code, "status_message": status_message, "response_message": response_message})
        else:
            return json.dumps({"status_code": status_code, "status_message": status_message})

    def handle_incoming_message(self, socket, message):
        return self.create_response_message(200, "OK", None)


class ZmqReceiverThread(Thread):
    def __init__(self, zmq_rep_bind_address=None, zmq_sub_connect_addresses=None, recreate_sockets_on_timeout_of_sec=60, username=None, password=None):
        Thread.__init__(self)
        self.receiver = ZmqReceiver(zmq_rep_bind_address=zmq_rep_bind_address, zmq_sub_connect_addresses=zmq_sub_connect_addresses, recreate_sockets_on_timeout_of_sec=recreate_sockets_on_timeout_of_sec, username=username, password=password)

    def last_received_message(self):
        return self.receiver.last_received_message

    def run(self):
        self.receiver.run()

    def stop(self):
        self.receiver.stop()



'''
Created on Apr 8, 2014

@author: Jan Verhoeven

@copyright: MIT license, see http://opensource.org/licenses/MIT
'''
import zmq
import json
import logging
import time

logger = logging.getLogger("zmqrpc")


# ZmqSender implements a ZeroMQ REQ or PUB socket to send messages out via a 
# send function. The send function is equipped with a timeout and automatic
# recreation of the underlying REQ socket if no message is received back in the
# given timeout.
# The username/password can be used to provide 'simple' protection on the wire
# (only PLAIN has been implemented, so be aware of sniffers).
class ZmqSender():
    def __init__(self, zmq_req_endpoints=None, zmq_pub_endpoint=None, username=None, password=None):
        self.context = zmq.Context()
        self.username = username
        self.password = password
        self.poller = zmq.Poller()
        self.zmq_req_endpoints = zmq_req_endpoints
        self.zmq_pub_endpoint = zmq_pub_endpoint
        self.pub_socket = None
        self.req_socket = None
        self.recreate_pub_socket = False
        self.recreate_req_socket = False
        self.create_pub_socket()
        self.create_req_socket()
        # 'Prevent' slow joiner problem
        time.sleep(0.5)

    def destroy_req_socket(self):
        error_message = None
        if self.req_socket is not None:
            try:
                self.poller.unregister(self.req_socket)
            except Exception as e:
                error_message = "Cannot unregister REQ socket to poller. Exception: {0}".format(e)
            try:
                self.req_socket.setsockopt(zmq.LINGER, 0)
            except Exception as e:
                error_message = "Cannot set LINGER socket option. Exception: {0}".format(e)
            try:
                for endpoint in self.zmq_req_endpoints:
                    self.req_socket.disconnect(endpoint)
            except Exception as e:
                error_message = "Cannot disconnect REQ socket. Exception: {0}".format(e)
            try:
                self.req_socket.close()
            except Exception as e:
                error_message = "Cannot close REQ socket. Exception: {0}".format(e)
        self.req_socket = None
        if error_message is not None:
            logger.error(error_message)

    def destroy_pub_socket(self):
        error_message = None
        if self.pub_socket is not None:
            try:
                self.pub_socket.setsockopt(zmq.LINGER, 0)
            except Exception as e:
                error_message = "Cannot set LINGER socket option. Exception: {0}".format(e)
            try:
                self.pub_socket.unbind(self.zmq_pub_endpoint)
            except Exception as e:
                raise Exception("Cannot unbind PUB socket from {0}. Exception: {1}".format(self.zmq_pub_endpoint, e))
            try:
                self.pub_socket.close()
            except Exception as e:
                error_message = "Cannot close PUB socket. Exception: {0}".format(e)
        self.pub_socket = None
        if error_message is not None:
            logger.error(error_message)

    def create_req_socket(self):
        if self.req_socket is not None:
            raise "Want create new REQ socket, but old REQ Socket is not destroyed."

        if self.zmq_req_endpoints:
            self.req_socket = self.context.socket(zmq.REQ)
            if self.username and self.password:
                self.req_socket.plain_username = self.username
                self.req_socket.plain_password = self.password
            try:
                for endpoint in self.zmq_req_endpoints:
                    self.req_socket.connect(endpoint)
            except Exception as e:
                raise Exception("Cannot connect REQ socket to {0}. Exception: {1}".format(self.zmq_req_endpoints, e))
            try:
                self.poller.register(self.req_socket, zmq.POLLIN)
            except Exception as e:
                raise Exception("Cannot register REQ socket to poller. Exception: {0}".format(e))

    def create_pub_socket(self):
        if self.pub_socket is not None:
            raise "Want create new PUB socket, but old PUB Socket is not destroyed."

        if self.zmq_pub_endpoint:
            self.pub_socket = self.context.socket(zmq.PUB)
            if self.username and self.password:
                self.pub_socket.plain_username = self.username
                self.pub_socket.plain_password = self.password
            # Make sure we buffer enough (100000 messages) in case of network troubles...
            self.pub_socket.setsockopt(zmq.SNDHWM, 100000)
            try:
                self.pub_socket.bind(self.zmq_pub_endpoint)
            except Exception as e:
                raise Exception("Cannot bind PUB socket to {0}. Exception: {1}".format(self.zmq_pub_endpoint, e))

    def _send_over_pub_socket(self, message):
        if self.pub_socket is not None:
            try:
                self.pub_socket.send(message)
            except Exception as e:
                self.recreate_pub_socket = True
                raise Exception("Cannot send message on PUB socket. Highly exceptional. Mark PUB socket for renewal. Consider this message lost. Exception: {0}".format(e))

    def handle_response(self, response_message_json):
        try:
            response_message_dict = json.loads(response_message_json)
        except:
            raise Exception("Marshalling error: Response is not a json message")
        else:
            if "status_code" in response_message_dict:
                if response_message_dict["status_code"] == 200 and "response_message" in response_message_dict:
                    return response_message_dict["response_message"]
                elif response_message_dict["status_code"] != 200 and "status_message" in response_message_dict:
                    raise Exception(response_message_dict["status_message"])
                elif response_message_dict["status_code"] != 200 and "status_message" not in response_message_dict:
                    raise Exception("Error occured with code {0}".format(response_message_dict["status_code"]))
            else:
                raise Exception("No status_code in response")

    def _send_over_req_socket(self, message, time_out_waiting_for_response_in_sec=10):
        if self.req_socket is not None:
            try:
                self.req_socket.send(message)
            except Exception as e:
                self.recreate_req_socket = True
                raise Exception("Cannot send message on REQ socket. This is very exceptional. Please check logs. Marking REQ socket to be recreated on next try. Message can be considered lost. Exception: {0}".format(e))
            else:
                # Wait for given time to receive response.
                start_time = time.time()
                while start_time + time_out_waiting_for_response_in_sec > time.time():
                    # X seconds timeout before quiting on waiting for response
                    req_socks = dict(self.poller.poll(1000))
                    if req_socks.get(self.req_socket) == zmq.POLLIN:
                        try:
                            response_message_json = self.req_socket.recv()
                        except Exception as e:
                            logger.error("Could not receive message from socket. Marking REQ socket to be recreated on next try. Exception: {0}".format(e))
                            self.recreate_req_socket = True
                        else:
                            return self.handle_response(response_message_json)
                # Some unexpected socket related error occurred. Recreate the REQ socket.
                self.recreate_req_socket = True
                raise Exception("No response received on ZMQ Request to end point {0} in {1} seconds. Discarding message. Marking REQ socket to be recreated on next try.".format(self.zmq_req_endpoints, time_out_waiting_for_response_in_sec))

    def send(self, message, time_out_waiting_for_response_in_sec=60):
        # Create sockets if needed. Raise an exception if any problems are encountered
        if self.recreate_pub_socket:
            self.destroy_pub_socket()
            self.create_pub_socket()
            self.recreate_pub_socket = False

        if self.recreate_req_socket:
            self.destroy_req_socket()
            self.create_req_socket()
            self.recreate_req_socket = False

        # Sockets must exist otherwise we would not be here...
        # Any errors in the following lines will throw an error that must be catched
        self._send_over_pub_socket(message)
        return self._send_over_req_socket(message, time_out_waiting_for_response_in_sec)

    def destroy(self):
        self.destroy_req_socket()
        self.destroy_pub_socket()
